/*
 * Copyright (C) 1990-1997 by CERN/CN/SW/DC
 * All rights reserved
 */

#ifndef lint
static char sccsid[] = "@(#)remote.c	1.8 08/22/97 CERN CN-SW/DC Felix Hassine";
#endif /* not lint */

#include <stdio.h>
#include <log.h> 
#include <sys/types.h>
#if defined(_WIN32)
#include <winsock2.h>
#define MAXHOSTNAMELEN 64
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/param.h>
#endif
#include <string.h>
#if defined(SOLARIS)
#include <sys/sockio.h>
#endif
#if defined(_AIX)
#include <sys/time.h>                   /* needed for if.h              */
#endif /* AIX */
#if !defined(_WIN32)
#include <net/if.h>                     /* Network interfaces           */
#include <sys/ioctl.h>                  /* ioctl() definitions          */
#endif

#ifndef LOCALHOSTSFILE
#define LOCALHOSTSFILE "/etc/shift.localhosts"
#endif
#ifndef RTHOSTSFILE
#define RTHOSTSFILE "/etc/shift.rthosts"
#endif

extern int errno;
extern char * sys_errlist[];

/*
 * isremote(): returns 0 if requestor is in site
 *                   1 if requestor is out of site
 *		    -1 in case of an error
 */
int     isremote(from_host,hostname)
struct in_addr from_host;
char *hostname ;
{
#if defined(_WIN32)
	return 0;		/* We don't know yet how to get the info */
#else
	char *p ;
        char local[MAXHOSTNAMELEN];
	char    buf[BUFSIZ];            /* A buffer                     */
	char ent[25] ;
        struct hostent  *h;
	int s_s ; 			/* Socket */
	int n ;			
	struct  ifconf  ifc;		/* ifconf structure      */
	struct  ifreq   *ifr;		/* Pointer on ifreq structure */
	long binaddr ;
	unsigned int netw ;
        union adr {
                u_long adr_i;
                unsigned char adr_c[4];
        } adr,*ladd;
	char    *endp;
	struct  sockaddr_in     *sp;
	struct in_addr in ;
	struct  sockaddr_in addr;
	extern char * getconfent() ;

	if ( (p=(char *)strtok(getconfent("SIMULATION","REMOTE",1)," \t"))!=NULL && !strcmp(p,"YES")) {
		log(LOG_DEBUG,"isremote(): Client simulates remote behaviour\n");
		return 1 ;
	}
	if ( (p=(char *)strtok(getconfent("ISREMOTE","CALLS",1)," \t") )!=NULL && !strcmp(p,"NO") ) {
                log(LOG_DEBUG,"isremote(): Any connection assumed from local site\n");
                return 0 ;
        }

	/*
	 * getting local IP number
	 */
	gethostname(local,MAXHOSTNAMELEN);
        if ( (h=(struct hostent *)gethostbyname(local))==NULL) {
                log(LOG_ERR,"isremote(): gethostbyname() error\n");
                return -1 ;
        }
        ladd=(union adr *)h->h_addr_list[0];
        in.s_addr=ladd->adr_i ;
        log(LOG_DEBUG, "isremote(): Local host is %s\n",inet_ntoa( in ));


	if ( hostname != NULL ) {
		FILE *fs;
		char *cp ;
		char s[MAXHOSTNAMELEN+1];


		/*
                 * Is the hostname declared as a "remote" site host ?
                 */
		log(LOG_DEBUG,"isremote(): searching <%s> in %s\n",hostname,RTHOSTSFILE);
		if ( (fs = fopen( RTHOSTSFILE,"r")) != NULL ) {
                        while ( fgets(s,MAXHOSTNAMELEN+1,fs) != NULL )
			  if ( (cp= strtok(s," \n\t"))!=NULL ) {
                                if ( !isdigit(cp[0]) && (cp[0]!='#') &&  !strcmp(cp,hostname) ) {
                                                log(LOG_DEBUG,"isremote(): %s is in list of external hosts\n",cp);
                                                fclose(fs);
                                                return 1;
                                }
				if ( isdigit(cp[0]) ) {
                                        strcpy(ent,cp) ;
					if ( strtok(cp,".") ==  NULL ||
                                             strtok(NULL,".") == NULL )
                                                log(LOG_DEBUG,"%s ignored: IP specification too short\n", ent);
                                        else {
                                                if ( !strncmp( ent, inet_ntoa( from_host ), strlen(ent))) {
                                                        log(LOG_DEBUG,"Entry %s matches to %s\n",ent,inet_ntoa(from_host));
                                                        log(LOG_INFO,"isremote(): %s is classified as remote\n",hostname);
							fclose(fs) ;
                                                        return 1 ;
                                                }
                                        }
				}
                          }
                        fclose(fs);
		}


                /*
                 * Is the hostname declared local ?
                 */
		log(LOG_DEBUG,"isremote(): searching <%s> in %s\n",hostname,LOCALHOSTSFILE);
		if ( (fs = fopen( LOCALHOSTSFILE,"r")) != NULL ) {
			while ( fgets(s,MAXHOSTNAMELEN+1,fs) != NULL )
			  if ( (cp= strtok(s," \n\t")) != NULL ) {
				if ( !isdigit(cp[0]) && (cp[0]!='#') &&  !strcmp(cp,hostname) ) {
					log(LOG_DEBUG,"isremote(): %s is in list of local hosts\n",cp);
					fclose(fs);
					return 0;
				}
				if ( isdigit(cp[0]) ) {

					strcpy(ent,cp) ;
					if ( strtok(cp,".") ==  NULL || 
					     strtok(NULL,".") == NULL )
						log(LOG_DEBUG,"%s ignored: IP specification too short \n", ent);
					else {
						if ( !strncmp( ent, inet_ntoa( from_host ), strlen(ent) )) {
							log(LOG_DEBUG,"Entry %s matches to %s\n",ent,inet_ntoa(from_host));
							log(LOG_DEBUG,"isremote(): %s is classified as local\n",hostname);
							fclose (fs);
							return 0 ;
						}
					}
				}
			   }
			fclose(fs);
		}
	}

	netw = inet_netof(from_host) ;
        adr.adr_i=from_host.s_addr;
	log(LOG_DEBUG, "isremote(): Client host is %s\n",inet_ntoa( from_host )) ;

        if ((s_s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
                log(LOG_ERR, "socket: %s\n",sys_errlist[errno]);
                return -1;
        }

	ifc.ifc_len = sizeof(buf);
        ifc.ifc_buf = buf;
        ifr = ifc.ifc_req;

        if ((n = ioctl(s_s, SIOCGIFCONF, (char *)&ifc)) < 0) {
                log(LOG_ERR, "ioctl(SIOCGIFCONF): %s\n",sys_errlist[errno]);
		close(s_s);
                return -1;
        }
        else    {
#if !defined(_AIX)
                for (n = ifc.ifc_len/sizeof(struct ifreq); --n >= 0; ifr++){
                        memcpy (&addr, &ifr->ifr_addr, sizeof(struct sockaddr_in));
			log(LOG_DEBUG , "Comparing %d and %d \n",  inet_netof(addr.sin_addr), inet_netof(from_host) ) ;
			if ( inet_netof(addr.sin_addr) == inet_netof(from_host) ) {
				close(s_s);
				log(LOG_DEBUG ,"isremote(): client is in same site\n");
				return 0;
			}
		}
#else
		
                endp = (char *) ifr + ifc.ifc_len;
                sp = (struct sockaddr_in *) &ifr->ifr_addr;
                while ((char *) sp < endp) {
			log(LOG_DEBUG , " Comparing %d and %d \n",netw, inet_netof(sp->sin_addr)) ;
                        if ((sp->sin_family == AF_INET) &&
                                ( netw == inet_netof (sp->sin_addr) )){
                                close(s_s);
				log(LOG_DEBUG ,"isremote(): client is in same site\n");
				return 0;
                        }
                        ifr = (struct ifreq *)((char *) sp + sp->sin_len);
                        sp = (struct sockaddr_in *) &ifr->ifr_addr;
                }
#endif

	}
	close(s_s);
        log(LOG_INFO ,"isremote(): client is in another site\n");
        return 1;
#endif
}
