/*
 * Copyright (C) 1997 by CERN/IT/PDP/IP
 * All rights reserved
 */
 
#ifndef lint
static char sccsid[] = "@(#)syslog.c	1.1 08/14/97 CERN IT-PDP/IP Christoph von Praun";
#endif /* not lint */

/* windows headers */
#include <sys/types.h>
#include <string.h>
#include <stdarg.h>   /* ANSI C*/
#include <time.h>
#include <stdio.h>
#include <process.h>
#include <stdlib.h>

/* windows headers */
#include <windows.h>
#include <wtypes.h>
#include <winbase.h>
#include <winerror.h>

/* local includes */
#include "syslog.h"
#include "slogmsg.h"


static HANDLE h;
static char c;

void syslog(int pri, const char *fmt, ...){
	va_list ap;
	va_start(ap, fmt);
	vsyslog(pri, fmt, ap);
	va_end(ap);
}
	
void vsyslog(int pri, const char *fmt, va_list ap)
{

	char *aszMsg[1], *p;
	WORD nt_prio;
	SID nt_sid;
	SID_IDENTIFIER_AUTHORITY nt_sid_authority;
	int saved_errno, nt_pid = _getpid();
	char tbuf[2048], fmt_cpy[1024];
	DWORD bufsize;
	BOOL result;
	char username[256];

	saved_errno = errno;

	/* Username retrieval */
	result = GetUserName(username, &bufsize);
		if (result == 0) {
		SYSLOG_DEBUG(printf("failed to initialise username\n"));
		SYSLOG_PANIC();
	}

	/* SID retrieval seems work this way ? */
	if (!InitializeSid(&nt_sid, &nt_sid_authority, 0)) {
		SYSLOG_DEBUG(printf("failed to initialise SID\n"));
		SYSLOG_PANIC();
	}

	(void)sprintf(tbuf, "[pid %d / user %s] ", nt_pid, username);
	for (p = tbuf; *p; ++p);

/********************************************
		 build the message

	/* substitute error message for %m */
	{
		char ch, *t1, *t2;
		char *strerror();

		for (t1 = fmt_cpy; ch = *fmt; ++fmt)
			if (ch == '%' && fmt[1] == 'm') {
				++fmt;
				for (t2 = strerror(saved_errno);
						*t1 = *t2++; ++t1);
			}
			else
				*t1++ = ch;
		*t1 = '\0';
	}

	(void)vsprintf(p, fmt_cpy, ap);

	aszMsg[0] = tbuf;

/*
	consider the EventTypes:
	EVENTLOG_ERROR_TYPE 
	EVENTLOG_WARNING_TYPE 
	EVENTLOG_INFORMATION_TYPE

/*******************************************************/

	switch(pri) {
	case LOG_EMERG:												/* system is unusable */
	case LOG_ALERT:												/* action must be taken immediately */
	case LOG_CRIT:												/* critical conditions */
	case LOG_ERR:													/* error conditions */
		nt_prio = EVENTLOG_ERROR_TYPE;
		break;
	case LOG_WARNING:											/* warning conditions */
	case LOG_NOTICE:											/* normal but significant condition */
		nt_prio = EVENTLOG_WARNING_TYPE;
		break;
	case LOG_INFO:												/* informational */
	case LOG_DEBUG:												/* debug-level messages */
		nt_prio = EVENTLOG_INFORMATION_TYPE;
		break;
	}

 
	if (!ReportEvent(
				h,										/* event log handle            */ 
        nt_prio,							/* event type                  */ 
        SYSLOG_ONE,           /* category identifier         */ 
        SYSLOG_01,            /* event identifier            */ 
        (PSID)&nt_sid,        /* no user security identifier */ 
        1,                    /* one substitution string     */ 
        0,                    /* no data                     */ 
        (LPTSTR *) aszMsg,    /* address of string array     */ 
        NULL)                 /* address of data             */ 
		) {
			SYSLOG_DEBUG(printf("could not report event; maybe log was not opened before?\n"));
			SYSLOG_PANIC(); 
		}
}



void openlog(const char *ident, int logstat, int logfac)
{
	HKEY hk; 
	DWORD dwData; 
	UCHAR szBuf[80];
	DWORD catCount;
	
	/* 
	 * Add your source name as a subkey under the Application 
	 * key in the EventLog service portion of the registry. 
	 */ 

	if (RegCreateKey(HKEY_LOCAL_MACHINE, SYSLOG_APP_REG_PATH, &hk)) {
			SYSLOG_DEBUG(printf("could not create nor open registry key"));
			SYSLOG_PANIC();
	}
 
	/* Set the Event ID message-file name. */ 
 
	strcpy(szBuf, SYSLOG_APP_FILE_PATH); 
 
	/* Add the Event ID message-file name to the subkey. */ 
	if (RegSetValueEx(hk,             /* subkey handle         */ 
					"EventMessageFile",       /* value name            */ 
					0,                        /* must be zero          */ 
					REG_EXPAND_SZ,            /* value type            */ 
					(LPBYTE) szBuf,           /* address of value data */ 
					strlen(szBuf) + 1)				/* length of value data  */ 
			) {      
			SYSLOG_DEBUG(printf("could not set event message file\n"));
			SYSLOG_PANIC();
	}

	/* Add the Category ID category-file name to the subkey. */ 
	strcpy(szBuf, SYSLOG_CAT_FILE_PATH); 
	if (RegSetValueEx(hk,             /* subkey handle         */ 
					"CategoryMessageFile",    /* value name            */ 
					0,                        /* must be zero          */ 
					REG_EXPAND_SZ,            /* value type            */ 
					(LPBYTE) szBuf,           /* address of value data */ 
					strlen(szBuf) + 1)				/* length of value data  */ 
			) {      
			SYSLOG_DEBUG(printf("could not set category message file\n"));
			SYSLOG_PANIC();
	}

	/* Set the category Counter */
	catCount = SYSLOG_CAT_COUNT;
	if (RegSetValueEx(hk,             /* subkey handle         */ 
					"CategoryCount",          /* value name            */ 
					0,                        /* must be zero          */ 
					REG_DWORD,                /* value type            */ 
					(LPBYTE) &catCount,       /* address of value data */ 
					sizeof(DWORD))				    /* length of value data  */ 
			) {      
			SYSLOG_DEBUG(printf("could not set category count file\n"));
			SYSLOG_PANIC();
	}


	/* Set the supported types flags. */ 
	dwData = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | EVENTLOG_INFORMATION_TYPE; 
 	if (RegSetValueEx(hk,      /* subkey handle                */ 
					"TypesSupported",  /* value name                   */ 
					0,                 /* must be zero                 */ 
					REG_DWORD,         /* value type                   */ 
					(LPBYTE) &dwData,  /* address of value data        */ 
					sizeof(DWORD))	   /* length of value data         */ 
			) {
		SYSLOG_DEBUG(printf("could not set supported types\n"));
		SYSLOG_PANIC(); 
	}

	RegCloseKey(hk);

	h = RegisterEventSource(NULL, /* uses local computer         */ 
			SYSLOG_APP_NAME);         /* source name                 */ 

	if (h == NULL) {	
		SYSLOG_DEBUG(printf("could not register event source\n"));
		SYSLOG_PANIC();
	}
}

void closelog() 
{
	DeregisterEventSource(h); 
};

